/*
 * Copyright 2012-14 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.internal.client.hdlc.physical;

import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Map;
import java.util.TooManyListenersException;

import org.openmuc.jdlms.client.hdlc.HdlcAddress;
import org.openmuc.jdlms.internal.client.communication.ILowerLayer;
import org.openmuc.jdlms.internal.client.communication.IUpperLayer;
import org.openmuc.jdlms.internal.client.hdlc.common.FrameInvalidException;
import org.openmuc.jdlms.internal.client.hdlc.common.HdlcAddressPair;
import org.openmuc.jdlms.internal.client.hdlc.common.HdlcHeaderParser;
import org.openmuc.jdlms.internal.util.ByteBufferInputStream;
import org.openmuc.jdlms.internal.util.LoggingHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class represents a connection on the physical layer according to IEC 62056-21 in protocol mode E
 * 
 * @author Karsten Mueller-Bier
 */
public class LocalDataExchangeClient implements ILowerLayer<HdlcAddressPair>, IPhysicalConnectionListener {

	private static Logger logger = LoggerFactory.getLogger(LocalDataExchangeClient.class);

	private static final byte[] REQUESTMESSAGE = new byte[] { (byte) 0x2F, (byte) 0x3F, (byte) 0x21, (byte) 0x0D,
			(byte) 0x0A };
	private static final byte[] ACKNOWLEDGE = new byte[] { (byte) 0x06, (byte) 0x32, (byte) 0x00, (byte) 0x32,
			(byte) 0x0D, (byte) 0x0A };

	private volatile boolean isConnected = false;
	private int connectedClients = 0;

	private PhysicalConnection connection = null;
	private final HdlcHeaderParser parser = new HdlcHeaderParser();
	private final Map<Object, IUpperLayer> listeners;

	private final String portName;

	private final int maxBaudrate;
	private final boolean useHandshake;

	private final ByteBuffer receivingDataBuffer = ByteBuffer.allocate(2048);

	public LocalDataExchangeClient(String portName, int baudrate, boolean useHandshake) {
		this.portName = portName;
		listeners = new HashMap<Object, IUpperLayer>(8);
		maxBaudrate = baudrate;
		this.useHandshake = useHandshake;
	}

	@Override
	public void connect(long timeout) throws IOException {
		if (!isConnected) {
			synchronized (this) {

				openPhysicalConnection();

				if (!useHandshake) {
					try {
						connection.setSerialParams(maxBaudrate, 8, 1, 0);
					} catch (UnsupportedCommOperationException e) {
						throw new IOException("Serial Port does not support " + maxBaudrate + "bd 8N1");
					}
					connection.startListening();
				}
				else {

					connection.send(REQUESTMESSAGE);

					int baudRateSetting = connection.listenForIdentificationMessage(timeout);

					int baudRate = getBaudRate(baudRateSetting);

					if (baudRate == -1) {
						throw new IOException(
								"Syntax error in identification message received: unknown baud rate received.");
					}

					ACKNOWLEDGE[2] = (byte) baudRateSetting;
					connection.send(ACKNOWLEDGE);

					// Sleep for about 250 milliseconds to make sure, that the
					// acknowledge message has been completely transmitted prior
					// to changing the baud rate
					try {
						Thread.sleep(250);
					} catch (InterruptedException e) {
					}

					try {
						connection.setSerialParams(baudRate, SerialPort.DATABITS_7, SerialPort.STOPBITS_1,
								SerialPort.PARITY_EVEN);
					} catch (UnsupportedCommOperationException e) {
						throw new IOException("Serial Port does not support " + baudRate + "bd 7E1");
					}
					connection.listenForAck(2000);

					try {
						connection.setSerialParams(baudRate, SerialPort.DATABITS_8, SerialPort.STOPBITS_1,
								SerialPort.PARITY_NONE);
					} catch (UnsupportedCommOperationException e) {
						throw new IOException("Serial Port does not support " + baudRate + "bd 8N1");
					}

					connection.startListening();

				}

				isConnected = true;
			}
		}
		connectedClients++;
	}

	@Override
	public void send(byte[] data) throws IOException {
		if (isConnected == false) {
			throw new IOException("Cannot send data. DLMS Client not connected.");
		}
		connection.send(data);
	}

	/**
	 * Additional note to {@code ITransportLayer#disconnect()}: Only disconnects the port if no other connection is
	 * using it
	 */
	@Override
	public void disconnect() throws IOException {
		connectedClients--;
		if (connectedClients <= 0) {
			isConnected = false;
			connection.removeListener();
			connection.close();
		}
	}

	@Override
	public void registerReceivingListener(HdlcAddressPair key, IUpperLayer listener) throws IllegalArgumentException {
		if (listeners.containsKey(key)) {
			throw new IllegalArgumentException("A connection with the addresses " + key.toString()
					+ " is already registered");
		}

		listeners.put(key, listener);
	}

	@Override
	public void removeReceivingListener(IUpperLayer listener) {
		while (listeners.values().remove(listener)) {
			; // Removes all references of listener
		}
	}

	@Override
	public void dataReceived(byte[] data, int length) {
		try {
			receivingDataBuffer.put(data, 0, length);

			ByteBuffer copy = ByteBuffer.wrap(receivingDataBuffer.array());
			copy.position(receivingDataBuffer.position());
			copy.flip();
			ByteBufferInputStream dataStream = new ByteBufferInputStream(copy);
			byte readByte = 0;
			while (dataStream.available() > 0 && readByte != 0x7E) {
				// Reads input stream until FLAG byte has been found
				readByte = (byte) dataStream.read();
			}
			while (dataStream.available() > 0) {

				do {
					copy.mark();
					readByte = (byte) dataStream.read();
				} while (readByte == 0x7E);
				copy.reset();

				byte[] frame = parser.readFrame(dataStream);
				HdlcAddressPair key = new HdlcAddressPair(parser.getLastValidDestination(), parser.getLastValidSource());

				if (dataStream.available() == 1) {
					dataStream.read(); // Read and discard final 0x7E flag
				}

				copy.compact();
				receivingDataBuffer.position(copy.position());
				copy.flip();

				if (HdlcAddress.isAllStation(parser.getLastValidSource())
						|| HdlcAddress.isNoStation(parser.getLastValidSource())) {
					// Source is not defined, discard
					logger.debug("Source is not defined. Frame discarded");
				}
				else if (listeners.containsKey(key)) {
					listeners.get(key).dataReceived(frame);
				}
			}

		} catch (IOException e) {
			// LoggingHelper.logStackTrace(e, logger);
		} catch (FrameInvalidException e) {
			LoggingHelper.logStackTrace(e, logger);
			ByteBuffer copy = ByteBuffer.wrap(receivingDataBuffer.array());
			copy.position(receivingDataBuffer.position());
			copy.flip();
			copy.get(); // Read over first 0x7E Flag

			// Read until end of buffer or next 0x7E Flag
			while (copy.remaining() > 0 && copy.get() != 0x7E) {
			}
			copy.compact();
			receivingDataBuffer.position(copy.position());
		}

	}

	@Override
	public void discardMessage(byte[] data) {
		// No messages are buffered on this layer
	}

	public boolean isConnected() {
		return isConnected;
	}

	/**
	 * Returns the baud rate chosen by the server for this communication
	 * 
	 * @param baudCharacter
	 *            Encoded baud rate (see 6.3.14 item 13c)
	 * @return The chosen baud rate or -1 on error
	 */
	private int getBaudRate(int baudCharacter) {
		int result = -1;
		switch (baudCharacter) {
		case 0x30:
			result = 300;
			break;
		case 0x31:
			result = 600;
			break;
		case 0x32:
			result = 1200;
			break;
		case 0x33:
			result = 2400;
			break;
		case 0x34:
			result = 4800;
			break;
		case 0x35:
			result = 9600;
			break;
		case 0x36:
			result = 19200;
			break;
		}
		return result;
	}

	private void openPhysicalConnection() throws IOException {
		if (connection == null || connection.isClosed()) {
			connection = acquireSerialPort(portName);
		}

		try {
			connection.registerListener(this);
		} catch (TooManyListenersException e) {
			throw new IOException("Port already in use", e);
		}
	}

	/**
	 * Tries to acquire the port named in the constructor
	 * 
	 * @throws NoSuchPortException
	 * @throws PortInUseException
	 * @throws IOException
	 * @throws UnsupportedCommOperationException
	 */
	public PhysicalConnection acquireSerialPort(String serialPortName) throws IOException {

		CommPortIdentifier portIdentifier;
		try {
			portIdentifier = CommPortIdentifier.getPortIdentifier(serialPortName);
		} catch (NoSuchPortException e) {
			throw new IOException("Serial port: " + serialPortName + " does not exist", e);
		}

		if (portIdentifier.isCurrentlyOwned()) {
			throw new IOException("Serial port " + serialPortName +  " is currently in use.");
		}

		CommPort commPort;
		try {
			commPort = portIdentifier.open(this.getClass().getName(), 2000);
		} catch (PortInUseException e) {
			throw new IOException("Serial port " + serialPortName + " is currently in use.", e);
		}

		if (!(commPort instanceof SerialPort)) {
			commPort.close();
			throw new IOException("The port " + serialPortName + " is not a serial port");
		}

		SerialPort serialPort;

		serialPort = (SerialPort) commPort;

		try {
			serialPort.setSerialPortParams(300, SerialPort.DATABITS_7, SerialPort.STOPBITS_1, SerialPort.PARITY_EVEN);
		} catch (UnsupportedCommOperationException e) {
			serialPort.close();
			serialPort = null;
			throw new IOException("Unable to set the baud rate or other serial port parameters", e);
		}

		PhysicalConnection result;
		try {
			result = new PhysicalConnection(serialPort);
		} catch (IOException e) {
			serialPort.close();
			serialPort = null;
			throw e;
		}

		return result;
	}
}
