/*
 * Copyright 2012-14 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.internal.client.hdlc.physical;

import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.TooManyListenersException;

import org.openmuc.jdlms.internal.util.LoggingHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Wrapper class around the actual SerialPort object, abstracting sending and receiving of data.
 * 
 * @author Karsten Mueller-Bier
 * @author Stefan Feuerhahn
 */
public class PhysicalConnection implements SerialPortEventListener {

	private static Logger logger = LoggerFactory.getLogger(PhysicalConnection.class);

	private final SerialPort serialPort;
	private IPhysicalConnectionListener listener = null;

	private boolean isClosed;

	private DataOutputStream os;
	private DataInputStream is;

	private static final int INPUT_BUFFER_LENGTH = 1024;
	private final byte[] buffer = new byte[INPUT_BUFFER_LENGTH];

	private static final int SLEEP_INTERVAL = 100;

	public PhysicalConnection(SerialPort serialPort) throws IOException {
		this.serialPort = serialPort;

		try {
			os = new DataOutputStream(serialPort.getOutputStream());
			is = new DataInputStream(serialPort.getInputStream());
		} catch (IOException e) {
			throw new IOException("Error getting input or output or input stream from serial port", e);
		}

		isClosed = false;
	}

	/**
	 * Sends the given bytes over the serial interface
	 * 
	 * @param data
	 *            Bytes to send
	 * @throws IOException
	 */
	public synchronized void send(byte[] data) throws IOException {
		LoggingHelper.logBytes(data, data.length, "Sending over " + serialPort.getName(), logger);
		os.write(data);
		os.flush();
	}

	/**
	 * Closes the serial interface, freeing the resource for other programs
	 */
	public synchronized void close() {
		if (isClosed == false) {
			serialPort.removeEventListener();
			serialPort.close();
			isClosed = true;
		}

	}

	/**
	 * Changes the connection parameter of the serial interface
	 * 
	 * @param baud
	 *            Baud rate to communicate
	 * @param databits
	 *            Number of Data bits (Range 7-8)
	 * @param stopbits
	 *            Number of Stop bits (Range 0-2)
	 * @param parity
	 *            Parity Bit (Range 0-2)
	 * @throws UnsupportedCommOperationException
	 */
	public void setSerialParams(int baud, int databits, int stopbits, int parity)
			throws UnsupportedCommOperationException {
		serialPort.setSerialPortParams(baud, databits, stopbits, parity);
		serialPort.enableReceiveTimeout(5);
	}

	/**
	 * Registers an upper layer as Listener for incoming data
	 * 
	 * @param listener
	 *            The listener to register
	 * @throws TooManyListenersException
	 *             If a listener has already been registered
	 */
	public void registerListener(IPhysicalConnectionListener listener) throws TooManyListenersException {
		if (this.listener != null) {
			throw new TooManyListenersException();
		}
		this.listener = listener;
	}

	/**
	 * Remove the actual listener from this serial interface
	 */
	public void removeListener() {
		listener = null;
	}

	/**
	 * Callback method when data is received from the wrapped SerialPort object
	 */
	@Override
	public void serialEvent(SerialPortEvent arg0) {
		int data;

		try {
			int len = 0;
			while ((data = serialPort.getInputStream().read()) > -1) {
				buffer[len++] = (byte) data;
			}
			LoggingHelper.logBytes(buffer, len, "Received from " + serialPort.getName(), logger);
			listener.dataReceived(buffer, len);
		} catch (IOException e) {
			LoggingHelper.logStackTrace(e, logger);
		}
	}

	public boolean isClosed() {
		return isClosed;
	}

	public int listenForIdentificationMessage(long timeout) throws IOException {

		boolean readSuccessful = false;
		int timeval = 0;
		int numBytesReadTotal = 0;

		while (timeout == 0 || timeval < timeout) {
			if (is.available() > 0) {

				int numBytesRead = is.read(buffer, numBytesReadTotal, INPUT_BUFFER_LENGTH - numBytesReadTotal);
				numBytesReadTotal += numBytesRead;

				if (numBytesRead > 0) {
					timeval = 0;
				}

				if ((numBytesReadTotal > 6) && buffer[numBytesReadTotal - 2] == 0x0D
						&& buffer[numBytesReadTotal - 1] == 0x0A) {
					readSuccessful = true;
					break;
				}
			}

			try {
				Thread.sleep(SLEEP_INTERVAL);
			} catch (InterruptedException e) {
			}

			timeval += SLEEP_INTERVAL;
		}

		if (!readSuccessful) {
			throw new IOException("Timeout while listening for Identification Message");
		}

		return buffer[4];
	}

	public void listenForAck(long timeout) throws IOException {
		int timeval = 0;
		int numBytesReadTotal = 0;

		while (timeval < timeout) {
			if (is.available() > 0) {
				int numBytesRead = is.read(buffer, numBytesReadTotal, INPUT_BUFFER_LENGTH - numBytesReadTotal);
				numBytesReadTotal += numBytesRead;

				if (numBytesRead > 0) {
					timeval = 0;
				}

				if ((numBytesReadTotal == 6)) {
					break;
				}
			}

			try {
				Thread.sleep(SLEEP_INTERVAL);
			} catch (InterruptedException e) {
			}

			timeval += SLEEP_INTERVAL;
		}

	}

	public void startListening() throws IOException {
		try {
			serialPort.addEventListener(this);
		} catch (TooManyListenersException e1) {
			throw new IOException("Too many listeners on serial port");
		}
		serialPort.notifyOnDataAvailable(true);
		try {
			serialPort.enableReceiveTimeout(35);
		} catch (UnsupportedCommOperationException e) {
			throw new IOException("unable to set serial port receive timeout");
		}
	}

}
